class YaTrackerException(Exception):
    pass


class NotAuthorized(YaTrackerException):
    def __init__(self):
        super().__init__('The user is not authorized. Check whether all the steps '
                         'described in API access were completed.')


class SufficientRights(YaTrackerException):
    def __init__(self):
        super().__init__('You do not have sufficient rights to perform this action. '
                         'Double-check permissions in the Tracker interface. You '
                         'need the same permissions to perform the action via the '
                         'API as in the interface.')


class ObjectNotFound(YaTrackerException):
    def __init__(self):
        super().__init__('The requested object was not found. You might have entered'
                         ' an incorrect object ID or key value.')


class AlreadyExists(YaTrackerException):
    def __init__(self):
        super().__init__('An issue with the same value of the unique parameter '
                         'already exists.')



from .exceptions import NotAuthorized, SufficientRights, ObjectNotFound
from .exceptions import YaTrackerException, AlreadyExists
from .issue import FullIssue
from .objects import Issue, IssueType, Priority, Queue, Sprint, Status, User
from .objects import Transitions, Transition, Comment



from typing import Optional, Union, Iterator

from .base import BaseObject


class Issue(BaseObject):
    url: str
    id: str
    key: str
    display: str

    class Config:
        fields = {
            'url': {'alias': '_self'},
        }


class User(BaseObject):
    url: str
    id: str
    display: str

    class Config:
        fields = {
            'url': {'alias': '_self'},
        }


class Sprint(BaseObject):
    url: str
    id: str
    display: str

    class Config:
        fields = {
            'url': {'alias': '_self'},
        }


class IssueType(BaseObject):
    url: str
    id: str
    key: str
    display: str

    class Config:
        fields = {
            'url': {'alias': '_self'},
        }


class Priority(BaseObject):
    """
    Attributes:
        url - Reference to the object.
        id - Priority ID.
        key - Priority key.
        version - Priority version.
        name - Display name of the priority. When localized=false is passed
                in the request, this parameter contains duplicates of
                the names in other languages.
        order - The weight of the priority. This parameter affects the order
                for displaying the priority in the interface.
    """
    url: str
    id: str
    key: str
    display: Optional[str]
    version: Optional[int]
    name: Optional[Union[str, dict]]
    order: Optional[int]

    class Config:
        fields = {
            'url': {'alias': '_self'},
        }


class Queue(BaseObject):
    url: str
    id: str
    key: str
    display: str

    class Config:
        fields = {
            'url': {'alias': '_self'},
        }


class Status(BaseObject):
    url: str
    id: str
    key: str
    display: str

    class Config:
        fields = {
            'url': {'alias': '_self'},
        }


class Transition(BaseObject):
    id: str
    url: str
    display: str
    to: Status

    class Config:
        fields = {
            'url': {'alias': '_self'},
        }

    async def execute(self):
        await self.tracker.execute_transition(self)


class Transitions(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current = -1

    def __iter__(self) -> Iterator[Transition]:
        return self

    def __next__(self):
        self.current += 1
        values = list(self.values())
        if self.current >= len(values):
            raise StopIteration
        return values[self.current]


class Comment(BaseObject):
    url: str
    id: str
    text: str
    created_by: User
    updated_by: User
    created_at: str
    updated_at: str
    version: int

    class Config:
        fields = {
            'url': {'alias': '_self'},
            'updated_by': {'alias': 'updatedBy'},
            'created_at': {'alias': 'createdAt'},
            'created_by': {'alias': 'createdBy'},
            'updated_at': {'alias': 'updatedAt'},
        }



import contextvars
from typing import TypeVar, Type

T = TypeVar('T')


class ContextInstanceMixin:
    def __init_subclass__(cls, **kwargs):
        cls.__context_instance = contextvars.ContextVar(f'instance_{cls.__name__}')
        return cls

    @classmethod
    def get_current(cls: Type[T], no_error=True) -> T:
        if no_error:
            return cls.__context_instance.get(None)
        return cls.__context_instance.get()

    @classmethod
    def set_current(cls: Type[T], value: T):
        if not isinstance(value, cls):
            raise TypeError(f'Value should be instance of {cls.__name__!r} not {type(value).__name__!r}')
        cls.__context_instance.set(value)



import asyncio as aio
import ssl
from typing import Optional, List

import certifi
import rapidjson
from aiohttp import ClientSession, TCPConnector, BasicAuth

from .types import FullIssue, Transitions, Transition, Priority, Comment, AlreadyExists
from .types import NotAuthorized, SufficientRights, ObjectNotFound, YaTrackerException
from .utils.mixins import ContextInstanceMixin


class YaTracker(ContextInstanceMixin):
    """
    API docs: https://tech.yandex.com/connect/tracker/api/about-docpage/

    Attention!
        All 'self' properties renamed to 'link' cause it's incompatible with Python.
        All camelCase properties renamed to pythonic_case.
        Methods named by author, cause Yandex API has no clear method names.
        For help you to recognize method names full description is attached.

    """
    host = 'https://api.tracker.yandex.net'
    _session: Optional[ClientSession] = None

    def __init__(self, org_id, token):
        self.__org_id = org_id
        self.__token = token

    async def get_issue(self, issue_id, expand=None):
        """
        View issue parameters.
        Use this request to get information about an issue.

        :type issue_id: str
        :param expand: Additional fields to include in the response:
                        transitions — Lifecycle transitions between statuses.
                        attachments — Attached files

        :return:
        """
        method = 'GET'
        uri = f'{self.host}/v2/issues/{issue_id}'
        params = {'expand': expand} if expand else None
        data = await self._request(method, uri, params)
        return FullIssue(**data)

    async def edit_issue(self, issue_id, version=None, **kwargs):
        """
        Edit an issue.
        Use this request to make changes to an issue.
        The issue is selected by its ID or key.

        :param kwargs:
        :param version:
        :param issue_id:
        :return:
        """
        method = 'PATCH'
        uri = f'{self.host}/v2/issues/{issue_id}'
        params = {'version': version} if version else None
        payload = self.clear_payload(kwargs)
        data = await self._request(method, uri, params, payload)
        return FullIssue(**data)

    async def create_issue(self, summary, queue, parent=None, description=None,
                           sprint=None, type=None, priority=None, followers=None,
                           unique=None, **kwargs):
        """
        Create an issue.
        Use this request to create an issue.

        :return:
        """
        payload = self.clear_payload(locals())
        method = 'POST'
        uri = f'{self.host}/v2/issues/'
        data = await self._request(method, uri, payload=payload)
        return FullIssue(**data)

    async def get_comments(self, issue_id):
        """
        Get the comments for an issue.
        Use this request to get a list of comments in the issue.
        :param issue_id:
        :return:
        """
        method = 'GET'
        uri = f'{self.host}/v2/issues/{issue_id}/comments'
        data = await self._request(method, uri)
        return [Comment(**item) for item in data]

    async def post_comment(self, issue_id, text, **kwargs):
        payload = self.clear_payload(locals(), exclude=['issue_id'])
        method = 'POST'
        uri = f'{self.host}/v2/issues/{issue_id}/comments/'
        data = await self._request(method, uri, payload=payload)
        return Comment(**data)

    async def count_issues(self, filter=None, query=None):
        """
        Get the number of issues.
        Use this request to find out how many issues meet the criteria in your request.
        :return:
        """
        method = 'POST'
        uri = f'{self.host}/v2/issues/_count'
        payload = dict()
        if filter:
            payload['filter'] = filter
        if query:
            payload['query'] = query

        data = await self._request(method, uri, payload=payload)
        return data

    async def find_issues(self, filter=None, query=None, order=None, expand=None,
                          keys=None, queue=None):
        """
        Find issues.
        Use this request to get a list of issues that meet specific criteria.
        If there are more than 10,000 issues in the response, use paging.
        :return:
        """
        if (filter or query) and (queue or keys):
            raise YaTrackerException('You can only use keys, queue or search request.')

        payload = self.clear_payload(locals(), exclude=['expand', 'order'])

        method = 'POST'
        uri = f'{self.host}/v2/issues/_search'

        params = dict()
        if order:
            params['order'] = order
        if expand:
            params['expand'] = expand

        data = await self._request(method, uri, params, payload)
        return [FullIssue(**item) for item in data]

    async def get_priorities(self, localized='True'):
        """
        Get priorities.
        Use this request to get a list of priorities for an issue.

        :param localized: 'True' or 'False'.
        :type localized: str
        :return:
        """
        method = 'GET'
        uri = f'{self.host}/v2/priorities'
        params = {'localized': localized} if localized else None
        data = await self._request(method, uri, params)
        return [Priority(**item) for item in data]

    async def get_issue_links(self, issue_id):
        """
        Get issue links.
        Use this request to get information about links between issues.
        The issue is selected by its ID or key.

        :param issue_id:
        :return:
        """
        method = 'GET'
        uri = f'{self.host}/v2/issues/{issue_id}/links'
        data = await self._request(method, uri)
        return [FullIssue(**item) for item in data]

    async def get_transitions(self, issue_id):
        """
        Get transitions.
        Use this request to get a list of possible transitions for an issue.
        The issue is selected by its ID or key.

        :param issue_id:
        :rtype: List[Transition]
        """
        method = 'GET'
        uri = f'{self.host}/v2/issues/{issue_id}/transitions'
        data = await self._request(method, uri)
        return Transitions(**{Transition(**item).id: Transition(**item) for item in data})

    async def execute_transition(self, transition: Transition, **kwargs):
        method = 'POST'
        uri = f'{transition.url}/_execute'
        payload = self.clear_payload(kwargs)
        data = await self._request(method, uri, payload)
        return [Transition(**item) for item in data]

    async def _request(self, method, uri, params=None, payload=None):
        """ Base request method. """

        # let's get new or existing session
        if not isinstance(self._session, ClientSession) or self._session.closed:
            await self._get_session()

        # set context
        YaTracker.set_current(self)

        # let's send request and get results
        async with self._session.request(method=method, url=uri,
                                         params=params, json=payload) as response:
            text = await response.text()
            self._check_status(response.status, text)
            return self._get_beauty_json(text)

    @staticmethod
    def _check_status(status, text):
        if status in (200, 201):
            return

        if status == 401:
            raise NotAuthorized()

        if status == 403:
            raise SufficientRights()

        if status == 404:
            raise ObjectNotFound()

        if status == 409:
            raise AlreadyExists()

        raise YaTrackerException(text)

    @staticmethod
    def clear_payload(payload: dict, exclude=None):
        exclude = exclude or []
        kwargs = payload.pop('kwargs', None)
        if kwargs:
            payload.update(kwargs)
        return {k: v for k, v in payload.items()
                if k not in ['self', 'cls'] + exclude and v is not None}

    @staticmethod
    def _get_beauty_json(text):
        """
        Ugly 'self' param escaping!
        Special thanks for Yandex API namespace incompatible with Python...

        """
        return rapidjson.loads(text.replace('"self":"', '"_self":"'))

    async def _get_session(self):
        """ Define aiohttp.ClientSession. """
        # set default headers
        headers = dict()
        headers['Authorization'] = f'OAuth {self.__token}'
        headers['X-Org-Id'] = f'{self.__org_id}'

        # add ssl support
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        connector = TCPConnector(ssl=ssl_context)

        # define session
        self._session = ClientSession(connector=connector, headers=headers,
                                      json_serialize=rapidjson.dumps)

    async def close(self):
        """ Graceful closing. """
        if isinstance(self._session, ClientSession) and not self._session.closed:
            await self._session.close()
            await aio.sleep(0.250)



#!/usr/bin/env python3
import pathlib
import re
import sys

from setuptools import find_packages, setup

try:
    from pip.req import parse_requirements
except ImportError:  # pip >= 10.0.0
    from pip._internal.req import parse_requirements

WORK_DIR = pathlib.Path(__file__).parent

# Check python version
MINIMAL_PY_VERSION = (3, 7)
if sys.version_info < MINIMAL_PY_VERSION:
    raise RuntimeError(
        "YaTracker works with Python {}+ only".format(".".join(map(str, MINIMAL_PY_VERSION)))
    )


def get_version():
    """
    Read version

    :return: str
    """
    txt = (WORK_DIR / "yatracker" / "__init__.py").read_text("utf-8")
    try:
        return re.findall(r"^__version__ = '([^']+)'\r?$", txt, re.M)[0]
    except IndexError:
        raise RuntimeError("Unable to determine version.")


def get_description():
    """
    Read full description from 'README.md'

    :return: description
    :rtype: str
    """
    with open("README.md", "r", encoding="utf-8") as f:
        return f.read()


def get_requirements(filename=None):
    """
    Read requirements from 'requirements txt'

    :return: requirements
    :rtype: list
    """
    if filename is None:
        filename = "requirements.txt"

    file = WORK_DIR / filename

    install_reqs = parse_requirements(str(file), session="hack")
    return [str(ir.req) for ir in install_reqs]


setup(
    name="yatracker",
    version=get_version(),
    packages=find_packages(exclude=("tests", "tests.*", "examples.*")),
    url="https://github.com/Olegt0rr/YaTracker",
    license="MIT",
    author="Oleg A.",
    requires_python=">=3.7",
    author_email="oleg@trueweb.app",
    description="Fully asynchronous library for Yandex Tracker",
    long_description=get_description(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
    install_requires=get_requirements(),
    package_data={"": ["requirements.txt"]},
    include_package_data=False,
)



from typing import List, Optional

from .base import BaseObject
from .objects import Issue, IssueType, Priority, Queue, Sprint, Status, User, Transitions


class FullIssue(BaseObject):
    url: str
    id: str
    key: str
    version: int

    summary: str
    parent: Optional[Issue]
    description: Optional[str]
    sprint: Optional[List[Sprint]]
    type: IssueType
    priority: Priority
    followers: Optional[List[User]]
    queue: Queue
    favorite: bool
    assignee: Optional[User]

    last_comment_update_at: Optional[str]
    aliases: Optional[List[str]]
    updated_by: Optional[User]
    created_at: str
    created_by: User
    votes: int
    updated_at: Optional[str]
    status: Status
    previous_status: Optional[Status]
    direction: Optional[str]

    class Config:
        extra = 'allow'
        fields = {
            'url': {'alias': '_self'},
            'last_comment_update_at': {'alias': 'lastCommentUpdatedAt'},
            'updated_by': {'alias': 'updatedBy'},
            'created_at': {'alias': 'createdAt'},
            'created_by': {'alias': 'createdBy'},
            'updated_at': {'alias': 'updatedAt'},
            'previous_status': {'alias': 'previousStatus'},
        }

    async def get_transitions(self) -> Transitions:
        """
        Returns dict and list-like Transitions object.

        Iterate Transitions like a list:
        >>> transitions = await self.get_transitions()
        >>> for t in transitions:
        >>>    print(t)

        Use Transitions like a dict with transition names:
        >>> transitions = await self.get_transitions()
        >>> close = transitions.get('close')
        >>> if close:
        >>>    await close.execute()
        """
        return await self.tracker.get_transitions(self.id)

    async def get_comments(self):
        """
        Get comments for self
        :return:
        """
        return await self.tracker.get_comments(self.id)

    async def post_comment(self, text=None, **kwargs):
        """
        Post comment for self
        :param text:
        :param kwargs:
        :return: Comment
        """
        return await self.tracker.post_comment(self.id, text=text, **kwargs)
